unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, Spin;

type
  TEditArray = array of TEdit;
  T2DEditArray = array of TEditArray;

  TFloatArray = array of Extended;
  T2DFloatArray = array of TFloatArray;

  TForm1 = class(TForm)
    LeftPanel: TPanel;
    ResultPanel: TPanel;
    RightPanel: TPanel;
    ControlsPanel: TPanel;
    SpinEdit1: TSpinEdit;
    Label1: TLabel;
    SolveButton: TButton;
    Bevel1: TBevel;
    procedure FormCreate(Sender: TObject);
    procedure CreateEdits(AOrder : integer);
    procedure DestroyEdits;
    procedure SpinEdit1Change(Sender: TObject);
    procedure SolveButtonClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    LeftEdits : T2DEditArray;
    RightEdits : TEditArray;
    ResultEdits : TEditArray;
    Order : integer;
  end;

function SolveLinearEquations(ALeftSide : T2DFloatArray ; ARightSide : TFloatArray) : TFloatArray;
function Determinant(ADet : T2DFloatArray) : Extended;

var
  Form1: TForm1;

implementation

{$R *.dfm}

// set initial order to 2
procedure TForm1.FormCreate(Sender: TObject);
begin
  Order := 2;
  CreateEdits(Order);
end;

// change the order of the equations
procedure TForm1.SpinEdit1Change(Sender: TObject);
begin
  if SpinEdit1.Text = '' then Exit;
  Order := SpinEdit1.Value;
  DestroyEdits;
  CreateEdits(Order);
end;

// create a grid of edit boxes representing the terms of the equations
procedure TForm1.CreateEdits(AOrder : integer);
Var
  LCol, LRow, Lx, Ly : integer;
begin
  Lx := 5;
  Ly := 5;
  LeftPanel.Width := AOrder * 32 + 9;
  LeftPanel.Height := AOrder * 22 + 9;
  RightPanel.Width := 41;
  RightPanel.Height := LeftPanel.Height;
  RightPanel.Left := LeftPanel.Width + 15;
  ResultPanel.Width := LeftPanel.Width;
  ResultPanel.Height := 31;
  ResultPanel.Top := LeftPanel.Top + LeftPanel.Height + 15;

  SetLength(LeftEdits, AOrder);
  SetLength(RightEdits, AOrder);

  for LRow := 0 to AOrder - 1 do begin

    // create rows of left side of equations
    SetLength(LeftEdits[LRow], AOrder);
    for LCol := 0 to AOrder - 1 do begin
      LeftEdits[LRow][LCol] := TEdit.Create(Self);
      with LeftEdits[LRow][LCol] do begin
        Parent := LeftPanel;
        Left := Lx;
        Top := Ly;
        Height := 20;
        Width := 30;
        Text := '';
      end;
      Inc(Lx, 32);
    end;

    // create right side of equations
    RightEdits[LRow] := TEdit.Create(Self);
    with RightEdits[LRow] do begin
      Parent := RightPanel;
      Left := 5;
      Top := Ly;
      Height := 20;
      Width := 30;
      Text := '';
    end;

    Lx := 5;
    Inc(Ly, 22);
  end;
end;

// destroy the edit boxes
procedure TForm1.DestroyEdits;
Var
  LCol, LRow : integer;
begin
  for LRow := 0 to High(LeftEdits) do begin
    // destroy left side of equations
    for LCol := 0 to High(LeftEdits[0]) do LeftEdits[LRow][LCol].Free;

    // destroy right side of equations
    RightEdits[LRow].Free;
  end;

  if length(ResultEdits) > 0 then begin
    // destroy results
    for LCol := 0 to High(ResultEdits) do ResultEdits[LCol].Free;
    SetLength(ResultEdits, 0);
  end;
end;

// solve the equations
procedure TForm1.SolveButtonClick(Sender: TObject);
Var
  LLeftSide : T2DFloatArray;
  LRightSide, LResults : TFloatArray;
  LRow, LCol : integer;
begin
  // transfer data from edits into extended arrays
  SetLength(LLeftSide, Order);
  SetLength(LRightSide, Order);
  for LRow := 0 to Order - 1 do begin
    SetLength(LLeftSide[LRow], Order);
    for LCol := 0 to Order - 1 do begin
      if LeftEdits[LRow][LCol].Text = '' then Exit;
      LLeftSide[LRow][LCol] := StrToFloat(LeftEdits[LRow][LCol].Text);
    end;
    if RightEdits[LRow].Text = '' then Exit;
    LRightSide[LRow] := StrToFloat(RightEdits[LRow].Text);
  end;

  // set dimensions of result array
  SetLength(LResults, Order);

  // solve the equations
  LResults := SolveLinearEquations(LLeftSide, LRightSide);

  // create result edit boxes
  LRow := 5;
  SetLength(ResultEdits, Order);
  for LCol := 0 to Order - 1 do begin
    ResultEdits[LCol] := TEdit.Create(Self);
    with ResultEdits[LCol] do begin
      Parent := ResultPanel;
      Left := LRow;
      Top := 5;
      Height := 20;
      Width := 30;
      Text := FloatToStr(LResults[LCol]);
      ReadOnly := True;
    end;
    Inc(LRow, 32);
  end;
end;

// solve a system of linear equations of any order using Cramer's method
function SolveLinearEquations(ALeftSide : T2DFloatArray ; ARightSide : TFloatArray) : TFloatArray;
Var
  LDenominator, LDenominatorInv : Extended;
  LOrder : integer;
  LRow, LCol : integer;
  LNumeratorDet : T2DFloatArray;
begin
  LOrder := length(ARightSide);
  SetLength(Result, LOrder);

  // special case - order = 1
  if LOrder = 1 then begin
    if ALeftSide[0][0] <> 0 then
      Result[0] := ARightSide[0] / ALeftSide[0][0] else
      ShowMessage('Equation is unsolvable - it is inconsistent');
    Exit;
  end;

  // solve the denominator of Cramer's equation
  // only needs to be calculated once
  LDenominator := Determinant(ALeftSide);
  
  if LDenominator = 0 then begin
    ShowMessage('Equations are unsolvable - they are either dependant or inconsistent');
    Exit;
  end;

  // calculate the inverse so that the divide is only done once
  LDenominatorInv := 1 / LDenominator;

  // create numerator for first variable
  SetLength(LNumeratorDet, LOrder);
  for LRow := 0 to LOrder - 1 do begin
    SetLength(LNumeratorDet[LRow], LOrder);
    LNumeratorDet[LRow][0] := ARightSide[LRow];
    for LCol := 1 to LOrder - 1 do begin
      LNumeratorDet[LRow][LCol] := ALeftSide[LRow][LCol];
    end;
  end;

  // solve the first variable
  Result[0] := Determinant(LNumeratorDet) * LDenominatorInv;

  // solve the remaining variables
  for LCol := 1 to LOrder - 1 do begin

    // restore the previous column with the left side values
    for LRow := 0 to LOrder - 1 do begin
      LNumeratorDet[LRow][LCol-1] := ALeftSide[LRow][LCol-1];
    end;

    // replace the current column with the right side values
    for LRow := 0 to LOrder - 1 do begin
      LNumeratorDet[LRow][LCol] := ARightSide[LRow];
    end;

    // solve for this variable
    Result[LCol] := Determinant(LNumeratorDet) * LDenominatorInv;
  end;
end;

// solve a determinant of any order
function Determinant(ADet : T2DFloatArray) : Extended;
Var
  LCol, LRow, LOrder, LNewOrder : Integer;
  LSign : boolean;
  LMinor : T2DFloatArray;
begin
  LOrder := length(ADet);

  if LOrder = 2 then begin

    // return the result
    Result := ADet[0][0] * ADet[1][1] - ADet[0][1] * ADet[1][0];

  end else begin

    // reduce the determinant by minors and solve determinants recursively
    LNewOrder := LOrder - 1;

    // create first minor
    SetLength(LMinor, LNewOrder);
    for LRow := 0 to LNewOrder - 1 do begin
      SetLength(LMinor[LRow], LNewOrder);
      for LCol := 0 to LNewOrder - 1 do
        LMinor[LRow][LCol] := ADet[LRow+1][LCol+1];
    end;

    // solve first minor
    Result := ADet[0][0] * Determinant(LMinor); 

    LSign := False; // negative

    // solve remaining minors and accumulate result
    for LCol := 1 to LNewOrder do begin

      // create the minor for this column by modifying the previous minor
      for LRow := 1 to LNewOrder do
        LMinor[LRow-1][LCol-1] := ADet[LRow][LCol-1];

      // accumulate terms to get result
      if LSign then Result := Result + ADet[0][LCol] * Determinant(LMinor)
               else Result := Result - ADet[0][LCol] * Determinant(LMinor);
               
      LSign := not LSign; // alternate sign of terms

    end; // minor columns loop

  end;
end;

end.
